package de.hsbo.fbv.bmg.topology.algorithms;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import de.hsbo.fbv.bmg.topology.networks.Edge;
import de.hsbo.fbv.bmg.topology.networks.Element;
import de.hsbo.fbv.bmg.topology.networks.Network1T;
import de.hsbo.fbv.bmg.topology.networks.Node;


/**
 * Diese Klasse stellt die wichtigsten Netzwerk-Algorithmen als statische Methoden bereit. 
 * @author A.Zimmermann, HS Bochum, 2012
 *
 */
public class Algorithms {
	
	// Parameter
	public static final int NONE = 0;
	public static final int WORK = 1;
	public static final int DONE = 2;
	public static final int CYCLE = 10;
	
	
	/**
	 * Fhrt die Breitensuche in einem Netzwerk durch.
	 * @param network Netzwerk
	 * @param start Startknoten
	 * @return
	 */
	public static synchronized Network1T runBreadthFirstSearch(Network1T network, Object start) {
		
		// Wenn Knoten start nicht zum Netzwerk gehrt -> Abbruch
		if(!network.containsNode(start))
			return null;
		
		// Sonst fhre Algorithmus durch
		// Bereite alle Knoten vor
		Node[] nodes = network.getNodes();
		prepareElements(nodes, Double.MAX_VALUE);

		// Setze Gewicht, Spur (Trace) und Status des Startknotens
		Node nstart = network.getNode(start);
		nstart.setWeight(0);
		nstart.setState(WORK);
		nstart.setTrace(nstart);
		
		// Richte Bearbeitungsliste ein
		List<Node> queue = new ArrayList<Node>();
		queue.add(nstart);
		
		// Durchlaufe Schleife solange, bis die Bearbeitungsliste leer ist
		do {
			// Entnehme den nchsten Knoten aus der Liste und setze Status auf DONE
			Node nwork = queue.remove(0);
			nwork.setState(DONE);
			// Suche dazu alle Kanten und bearbeite sie
			Edge[] edges = nwork.getEdges();
			for(int i=0; i<edges.length; i++) {
				// nchsten Knoten suchen
				Node nnext = edges[i].getOppositeNode(nwork);
				// Knoten wurde noch nicht besucht
				if(nnext.getState()==NONE) {
					nnext.setWeight(edges[i].getWeight()+nwork.getWeight());
					nnext.setTrace(edges[i]);
					nnext.setState(WORK);
					queue.add(nnext);
					
				}
				// Knoten ist in der aktuellen Bearbeitungsliste
				else if(nnext.getState()==WORK) {
					double weight = edges[i].getWeight()+nwork.getWeight();
					if(weight<nnext.getWeight()) {
						nnext.setWeight(weight);
						nnext.setTrace(edges[i]);
					}
				}
				// Knoten ist bereits vollstndig bearbeitet (DONE)
				else;
			}
		} while(!queue.isEmpty());

		// Ergebnis zurckgeben
		return network.getSubnet();
	}	// Ende runBreadthFirstSearch()
	
	/**
	 * Fhrt den Dijkstra-Algorithmus (krzeste Wege) durch.
	 * @param network Netzwerk
	 * @param start Startknoten
	 * @return
	 */
	public static synchronized Network1T runDijkstra(Network1T network, Object start) {
		
		// Wenn Knoten start nicht zum Netzwerk gehrt -> Abbruch
		if(!network.containsNode(start))
			return null;
		
		// Sonst fhre Algorithmus durch
		// Bereite alle Knoten vor
		Node[] nodes = network.getNodes();
		prepareElements(nodes, Double.MAX_VALUE);
		
		// Setze Gewicht, Spur (Trace) und Status des Startknotens
		Node nstart = network.getNode(start);
		nstart.setWeight(0);
		nstart.setState(WORK);
		nstart.setTrace(nstart);
		
		// Richte Bearbeitungsliste ein
		List<Node> queue = new ArrayList<Node>();
		queue.add(nstart);
		
		// Durchlaufe Schleife solange, bis die Bearbeitungsliste leer ist
		do {
			// Ergnzung fr den Dijkstra-Algorithmus: Suche in Liste queue
			// nach Knoten mit minimalem Gewicht und setzte ihn an Index 0
			getMinWeight(queue);
			// Entnehme den nchsten Knoten aus der Liste und setze Status auf DONE
			Node nwork = queue.remove(0);
			nwork.setState(DONE);
			// Suche dazu alle Kanten und bearbeite sie
			Edge[] edges = nwork.getEdges();
			for(int i=0; i<edges.length; i++) {
				// nchsten Knoten suchen
				Node nnext = edges[i].getOppositeNode(nwork);
				// Knoten wurde noch nicht besucht
				if(nnext.getState()==NONE) {
					nnext.setWeight(edges[i].getWeight()+nwork.getWeight());
					nnext.setTrace(edges[i]);
					nnext.setState(WORK);
					queue.add(nnext);
				}
				// Knoten ist in der aktuellen Bearbeitungsliste
				else if(nnext.getState()==WORK) {
					double weight = edges[i].getWeight()+nwork.getWeight();
					if(weight<nnext.getWeight()) {
						nnext.setWeight(weight);
						nnext.setTrace(edges[i]);
					}
				}
				// Knoten ist bereits vollstndig bearbeitet (DONE)
				else;
			}
		} while(!queue.isEmpty());

		// Ergebnis zurckgeben
		return network.getSubnet();
	}	// Ende runDijkstra()
	
	/**
	 * Fhrt den Dijkstra-Algorithmus zwischen Start- und Endknoten durch (krzeste Wege) durch.
	 * @param network Netzwerk
	 * @param start Startknoten
	 * @param end Endknoten
	 * @return
	 */
	public static synchronized Network1T runDijkstra(Network1T network, Object start, Object end) {
		
		// Wenn Knoten start nicht zum Netzwerk gehrt -> Abbruch
		if(!network.containsNode(start) || !network.containsNode(end))
			return null;
		
		// Sonst fhre Algorithmus durch
		// Bereite alle Knoten vor
		Node[] nodes = network.getNodes();
		prepareElements(nodes, Double.MAX_VALUE);
		
		// Setze Gewicht, Spur (Trace) und Status des Startknotens
		Node nstart = network.getNode(start);
		nstart.setWeight(0);
		nstart.setState(WORK);
		nstart.setTrace(nstart);
		// Setze Endknoten
		Node nend = network.getNode(end);
		
		// Richte Bearbeitungsliste ein
		List<Node> queue = new ArrayList<Node>();
		queue.add(nstart);
		
		// Durchlaufe Schleife solange, bis die Bearbeitungsliste leer ist
		do {
			// Ergnzung fr den Dijkstra-Algorithmus: Suche in Liste queue
			// nach Knoten mit minimalem Gewicht und setzte ihn an Index 0
			getMinWeight(queue);
			// Entnehme den nchsten Knoten aus der Liste und setze Status auf DONE
			Node nwork = queue.remove(0);
			nwork.setState(DONE);
			// Suche dazu alle Kanten und bearbeite sie
			Edge[] edges = nwork.getEdges();
			for(int i=0; i<edges.length; i++) {
				// nchsten Knoten suchen
				Node nnext = edges[i].getOppositeNode(nwork);
				// Knoten wurde noch nicht besucht
				if(nnext.getState()==NONE) {
					nnext.setWeight(edges[i].getWeight()+nwork.getWeight());
					nnext.setTrace(edges[i]);
					nnext.setState(WORK);
					queue.add(nnext);
				}
				// Knoten ist in der aktuellen Bearbeitungsliste
				else if(nnext.getState()==WORK) {
					double weight = edges[i].getWeight()+nwork.getWeight();
					if(weight<nnext.getWeight()) {
						nnext.setWeight(weight);
						nnext.setTrace(edges[i]);
					}
				}
				// Knoten ist bereits vollstndig bearbeitet (DONE)
				else;
			}
		} while(!queue.isEmpty() && nend.getState()!=DONE);

		// Ergebnis zurckgeben
		return network.getSubnet(end);
	}	// Ende runDijkstra()
	
	/**
	 * Fhrt die Tiefensuche in einem Netzwerk durch.
	 * @param network Netzwerk
	 * @return
	 */
	public static synchronized Network1T runDepthFirstSearch(Network1T network) {
		
		// Bereite alle Knoten vor
		Node[] nodes = network.getNodes();
		prepareElements(nodes, Double.MAX_VALUE);
		
		// Bereite alle Kanten vor
		Edge[] edges = network.getEdges();
		prepareElements(edges);
		
		// Alle Teilnetze besuchen 
		for(int i=0; i<nodes.length; i++) {
			if(nodes[i].getState()==NONE) {
				// Setzt den Wert des Anfangsknotens auf 0, den Status auf WORK 
				// und trace auf sich selbst
				Node nstart = nodes[i];
				nstart.setWeight(0);
				nstart.setState(WORK);
				nstart.setTrace(nstart);
				
				// Rekursiver Durchlauf des DFS-Algorithmus 
				visitDFS(nstart);
			}
		}
		
		// Ergebnis zurckgeben
		Network1T result = network.getSubnet(CYCLE);
		return result;
	}	// Ende runDeprhFirstSearch()

	/**
	 * Fhrt die Tiefensuche in einem Netzwerk durch.
	 * @param network Netzwerk
	 * @param start Startknoten
	 * @return
	 */
	public static synchronized Network1T runDepthFirstSearch(Network1T network, Object start) {
		// Prfung: Ist der Knoten mit dem Bezugsobjekt start im Netzwerk vorhanden?
		if(network.getNode(start)==null)
			return null;
		
		// Sonst fhre Algorithmus durch
		// Bereite alle Knoten vor
		Node[] nodes = network.getNodes();
		prepareElements(nodes, Double.MAX_VALUE);
		
		// Bereite alle Kanten vor
		Edge[] edges = network.getEdges();
		prepareElements(edges);
				
		// Setzt den Wert des Anfangsknotens auf 0, den Status auf WORK 
		// und trace auf sich selbst
		Node nstart = network.getNode(start);
		nstart.setWeight(0);
		nstart.setState(WORK);
		nstart.setTrace(nstart);
		
		// Rekursiver Durchlauf des DFS-Algorithmus 
		visitDFS(nstart);
		
		// Ergebnis zurckgeben
		Network1T result = network.getSubnet(CYCLE);
		return result;
	}	// Ende runDeprhFirstSearch()
	
	/**
	 * Interne Methode: Rekursiver Teil des Algorithmus zur Tiefensuche
	 * @param root
	 */
	private static void visitDFS(Node root) {
		// Alle Folgeknoten suchen und bearbeiten
		Edge[] edges = root.getEdges();
		// Schleife ber alle anhngenden Kanten
		for (int i=0; i<edges.length; i++) {
			// Kante und Folgeknoten suchen
			Edge edge = edges[i];
			Node node = edge.getOppositeNode(root);
			// Knoten und Kante wurden noch nicht besucht (NONE)
			if(node.getState()==NONE && edge.getState()==NONE) {
				node.setWeight(root.getWeight()+1);
				node.setTrace(root);
				node.setState(WORK);
				edge.setState(DONE);
				// rekursiver Aufruf
				visitDFS(node);
			}
			// Knoten-Status ist WORK oder DONE, Kanten-Staus ist NONE
			else if(edge.getState()==NONE) {
				edges[i].setState(CYCLE); 	// Zyklus gefunden
			}	
			// sonst
			else;	// keine Aktion
		}
		// Alle anhngenden Kanten (und Knoten) wurden besucht
		// Knoten als bearbeitet (Status DONE) kennzeichnen
		root.setState(DONE);
	}	// Ende visitDFS()

	/**
	 * Interne Methode: Setzt innerhalb einer Liste von Knoten denjenigen an 
	 * erste Stelle (Index = 0), der das kleinste Gewicht besitzt.
	 * @param list
	 */
	private static void getMinWeight(List<Node> list) {
		if(list.size()<=1)
			return;
		else {
			double weight = list.get(0).getWeight();
			for(int i=1; i<list.size(); i++) {
				if(list.get(i).getWeight()<weight) {
					list.add(0, list.remove(i));
					weight = list.get(0).getWeight();
				}
			}
		}	
	}	// Ende getMinWeight()
	
	/**
	 * Interne Methode: Belegt alle gegebenen Elemente mit dem Gewichtswert weight.
	 * @param elems
	 * @param weight
	 */
	private static void prepareElements(Element[] elems, double weight) {
		for(int i=0; i<elems.length; i++) {
			elems[i].setState(NONE);
			elems[i].setWeight(weight);
			elems[i].setTrace(null);
		}
	}	// Ende prepareAllNodes()

	/**
	 * Interne Methode: Belegt alle gegebenen Elemente mit dem Status = NONE und 
	 * einem leeren Vorgnger (trace = null).
	 * @param elems
	 */
	private static void prepareElements(Element[] elems) {
		for(int i=0; i<elems.length; i++) {
			elems[i].setState(NONE);
			elems[i].setTrace(null);
		}
	}	// Ende prepareAllNodes()

}
